/*
This file is part of Mitsuba, a physically based rendering system.

Copyright (c) 2007-2014 by Wenzel Jakob and others.

Mitsuba is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License Version 3
as published by the Free Software Foundation.

Mitsuba is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <mitsuba/render/bsdf.h>
#include <mitsuba/render/texture.h>
#include <mitsuba/hw/basicshader.h>
#include <mitsuba/core/warp.h>
//#include "furI.h"
#include "mitsuba/render/scene.h"
#include <vector>
#include <map>
//#include"furnet.h"
#include"aggregated.h"
#include"fur_utilize.h"
using namespace std;

MTS_NAMESPACE_BEGIN

class aggregated : public BSDF {
public:
	aggregated(const Properties &props)
		: BSDF(props) {
		// If color texture is given, read and convert it into absorption on the fly
		colorTexture = new ConstantSpectrumTexture(props.getSpectrum("color", Spectrum(1.0f)));
		useColorTexture = props.getBoolean("useColorTexture", false);
		medulla_ratio = props.getFloat("medulla_ratio", 0.0);//fur parameter
		eta = props.getFloat("eta", 1.55);//fur parameter
		beta_m = props.getFloat("beta_m", 0.055);//fur parameter
		beta_n = props.getFloat("beta_n", 0.055);//fur parameter
		alpha = props.getFloat("alpha", 3.0);//fur parameter
		g = props.getFloat("g", 0.0);//fur parameter
		layers = props.getFloat("layers", 0.5);//fur parameter
		absorptionFactor = props.getFloat("absorptionFactor", 4.0);//fur parameter
		absorption_outer = props.getVector("absorption_outer", Vector3(0.0f, 0.0f, 0.0f));//fur parameter
		absorption_inner = props.getFloat("absorption_inner", 0.0f);//fur parameter
		scattering_inner = props.getFloat("scattering_inner", 0.5f);//fur parameter
		furMaskCode = props.getInteger("furMaskCode", 31);//mask code for fur	
		// Render mode:
		//     "near":      near field only
		//     "far":       far field only
		//     "nearfar":   near field for first bounce, far field for others
		renderMode = props.getString("renderMode", "far");//in this implementation, we only consider far field
		// Color scale:
		//     > 0:             scale by colorScale
		// Color gamma:
		//     > 0:             gamma by colorScale
		colorScale = props.getFloat("colorScale", 2.0);//parameter to contorl absorption
		colorGamma = props.getFloat("colorGamma", 1.0);//parameter to contorl absorption
		// Tilt should always be negative for physical correctness.
		alpha = -degToRad(alpha);
		netinitialize("D://bin//furnet.bin");// initialize the small lightweight network,please change the path if you put the network other place
		initialize("medulla");//iinitialize the precomputed medulla, here please refer to the code published by [Yan et al. 2017] near-far filed.
	}

	aggregated(Stream *stream, InstanceManager *manager)
		: BSDF(stream, manager) {
		colorTexture = static_cast<Texture *>(manager->getInstance(stream));
		useColorTexture = stream->readBool();
		medulla_ratio = stream->readFloat();
		eta = stream->readFloat();
		beta_m = stream->readFloat();
		beta_n = stream->readFloat();
		alpha = stream->readFloat();
		g = stream->readFloat();
		layers = stream->readFloat();
		absorptionFactor = stream->readFloat();
		absorption_outer.x = stream->readFloat();
		absorption_outer.y = stream->readFloat();
		absorption_outer.z = stream->readFloat();
		absorption_inner = stream->readFloat();
		scattering_inner = stream->readFloat();
		renderMode = stream->readString();
		colorScale = stream->readFloat();
		colorGamma = stream->readFloat();
		netinitialize("D://onnx//furnet.bin");
		initialize("medulla");
		configure();
	}

	void configure() {
		/* Verify the input parameter and fix them if necessary */
		colorTexture = ensureEnergyConservation(colorTexture, "color", 1.0f);
		m_components.clear();
		m_components.push_back(EFUR2);
		if (colorTexture->getMaximum().max() > 0)
			m_components.push_back(EDiffuseReflection | EFrontSide
			| (colorTexture->isConstant() ? 0 : ESpatiallyVarying));
		m_usesRayDifferentials = colorTexture->usesRayDifferentials();
		BSDF::configure();
	}

	Spectrum getDiffuseReflectance(const Intersection &its) const {
		return colorTexture->eval(its);
	}

	Vector3 normalizeSafe(Vector3 v) const {
		if (v.length() == 0.0)
			return v;
		else
			return normalize(v);
	}

	Spectrum eval(const BSDFSamplingRecord &bRec, EMeasure measure) const {
		Vector3 absorption = absorption_outer;
		Float umin = bRec.its.storage[3];//the color area covered by an aggregated fiber
		Float umax = bRec.its.storage[4];//the color area covered by an aggregated fiber
		Float vmin = bRec.its.storage[5];//the color area covered by an aggregated fiber
		Float vmax = bRec.its.storage[6];//the color area covered by an aggregated fiber
		if (useColorTexture && medulla_ratio < 1) {
			Spectrum color = averagecolor(colorTexture,umin,umax,vmin,vmax);//average the absorption color
			Float color_r, color_g, color_b;//calculation follow [Yan et al. 2017]
			color.toLinearRGB(color_r, color_g, color_b);
			color_r = clamp(std::pow(color_r, Float(1.0 / colorGamma)) * colorScale, Float(1.0 / 256.0), Float(0.99));
			color_g = clamp(std::pow(color_g, Float(1.0 / colorGamma)) * colorScale, Float(1.0 / 256.0), Float(0.99));
			color_b = clamp(std::pow(color_b, Float(1.0 / colorGamma)) * colorScale, Float(1.0 / 256.0), Float(0.99));
			Float medulla_occlusion = std::pow(medulla_ratio, 1.5);
			Float absorption_r = (-log(color_r) - absorptionFactor * medulla_occlusion * absorption_inner) / (absorptionFactor - absorptionFactor * medulla_occlusion);
			Float absorption_g = (-log(color_g) - absorptionFactor * medulla_occlusion * absorption_inner) / (absorptionFactor - absorptionFactor * medulla_occlusion);
			Float absorption_b = (-log(color_b) - absorptionFactor * medulla_occlusion * absorption_inner) / (absorptionFactor - absorptionFactor * medulla_occlusion);
			absorption_r = clamp(absorption_r, 0.01, MAX_ABSORPTION_OUTER);
			absorption_g = clamp(absorption_g, 0.01, MAX_ABSORPTION_OUTER);
			absorption_b = clamp(absorption_b, 0.01, MAX_ABSORPTION_OUTER);
			absorption = Vector3(absorption_r, absorption_g, absorption_b);
		}
		//convert the w to theta and phi
		Vector3 wi = bRec.its.toWorld(bRec.wi);
		Vector3 wo = bRec.its.toWorld(bRec.wo);
		const Vector3 &u = bRec.its.geoFrame.s;
		const Vector3 &v = bRec.its.geoFrame.t;
		const Vector3 &w = bRec.its.geoFrame.n;
		Float theta_i = M_PI / 2.0f - acosSafe(dot(u, wi));
		Float theta_r = M_PI / 2.0f - acosSafe(dot(u, wo));
		//rolling the incident and outgoing direction on the sphere~~
		regularizetheta(theta_i, phi_i);
		regularizetheta(theta_r, phi_r);
		Float phi_i = acosSafe(dot(normalizeSafe(wi - dot(u, wi) * u), v)) * (dot(w, wi) > 0 ? 1 : -1);
		Float phi_r = acosSafe(dot(normalizeSafe(wo - dot(u, wo) * u), v)) * (dot(w, wo) > 0 ? 1 : -1);
		Float phi = phi_r - phi_i;
		regularizePhi(phi);
		Float rho = bRec.its.storage[7];//geometery parameter of the aggregated fiber
		int numoffibers = bRec.its.storage[8];//geometery parameter of the aggregated fiber
		AggregatedSolver AggregatedSolver(eta, absorption, alpha, beta_m, beta_n,
			medulla_ratio, absorption_inner, scattering_inner, g, layers, rho, numoffibers, theta_i, bRec.sampler);
		Vector3 R, TT, TRT, TTs, TRTs,top,middle,bottom,direct;	
		Float ms = AggregatedSolver.Ms(theta_i - alpha, theta_r - alpha, phi);
		//Inherited part
		R = AggregatedSolver.Mu(0, theta_i, theta_r, phi) * AggregatedSolver.Nu(0, phi, theta_i);
		TT = AggregatedSolver.Mu(1, theta_i, theta_r, phi) * AggregatedSolver.Nu(1, phi, theta_i);
		TRT = AggregatedSolver.Mu(2, theta_i, theta_r, phi) * AggregatedSolver.Nu(2, phi, theta_i);
		TTs = ms * max(discardLarger(AggregatedSolver.Ns(1, phi, theta_i), 10.0f), 0.0f);
		TRTs = ms * max(discardLarger(AggregatedSolver.Ns(2, phi, theta_i), 10.0f), 0.0f);
		//Extended part
		top = AggregatedSolver.Me(0, theta_i, theta_r, phi)*AggregatedSolver.Ne(0, theta_i, phi);
		middle = AggregatedSolver.Me(1, theta_i, theta_r, phi)*AggregatedSolver.Ne(1, theta_i, phi);
		bottom = AggregatedSolver.Me(2, theta_i, theta_r, phi)*AggregatedSolver.Ne(2, theta_i, phi);
		//Direct transport
		direct = AggregatedSolver.Md(theta_i, theta_r)*AggregatedSolver.Nd(phi);
		R = furMaskCode & 1 ? R : Vector3(Float(0));
		TT = (furMaskCode >> 1) & 1 ? TT : Vector3(Float(0));
		TRT = (furMaskCode >> 2) & 1 ? TRT : Vector3(Float(0));
		TTs = (furMaskCode >> 3) & 1 ? TTs : Vector3(Float(0));
		TRTs = (furMaskCode >> 4) & 1 ? TRTs : Vector3(Float(0));
		top = (furMaskCode >> 5) & 1 ? top : Vector3(Float(0));
		middle = (furMaskCode >> 6) & 1 ? middle : Vector3(Float(0));
		bottom = (furMaskCode >> 7) & 1 ? bottom : Vector3(Float(0));
		direct = (furMaskCode >> 8) & 1 ? direct : Vector3(Float(0));
		Vector3 bsdfCosine;
		bsdfCosine = (R + TT + TRT + TTs + TRTs+top+middle+bottom+direct) / cos(theta_r);
		//check invalid number
		if (std::isnan(bsdfCosine[0]) || std::isnan(bsdfCosine[1]) || std::isnan(bsdfCosine[2]))
			return Spectrum(0.0f);
		if (std::isinf(bsdfCosine[0]) || std::isinf(bsdfCosine[1]) || std::isinf(bsdfCosine[2]))
			return Spectrum(0.0f);
		for (int i = 0; i < 3; i++)
			bsdfCosine[i] = std::min(std::max(Float(0.0f), bsdfCosine[i]), Float(1e3f));
		Spectrum retVal(0.0f);
		retVal.fromLinearRGB(bsdfCosine[0], bsdfCosine[1], bsdfCosine[2]);
		return retVal;
	}

	Float normrnd(const BSDFSamplingRecord &bRec, Float mu, Float sigma) const {
		double U = bRec.sampler->next1D();
		double V = bRec.sampler->next1D();
		double X = sqrt(-2.0 * log(U)) * cos(2.0 * M_PI * V);
		return X * sigma + mu;
	}

	Float pdf(const BSDFSamplingRecord &bRec, EMeasure measure) const {
		Vector3 absorption = absorption_outer;
		Float umin = bRec.its.storage[3];//the color area covered by an aggregated fiber
		Float umax = bRec.its.storage[4];//the color area covered by an aggregated fiber
		Float vmin = bRec.its.storage[5];//the color area covered by an aggregated fiber
		Float vmax = bRec.its.storage[6];//the color area covered by an aggregated fiber
		if (useColorTexture && medulla_ratio < 1) {
			Spectrum color = averagecolor(colorTexture, umin, umax, vmin, vmax);//average the absorption color
			Float color_r, color_g, color_b;
			color.toLinearRGB(color_r, color_g, color_b);

			color_r = clamp(std::pow(color_r, Float(1.0 / colorGamma)) * colorScale, Float(1.0 / 256.0), Float(0.99));
			color_g = clamp(std::pow(color_g, Float(1.0 / colorGamma)) * colorScale, Float(1.0 / 256.0), Float(0.99));
			color_b = clamp(std::pow(color_b, Float(1.0 / colorGamma)) * colorScale, Float(1.0 / 256.0), Float(0.99));

			Float medulla_occlusion = std::pow(medulla_ratio, 1.5);
			Float absorption_r = (-log(color_r) - absorptionFactor * medulla_occlusion * absorption_inner) / (absorptionFactor - absorptionFactor * medulla_occlusion);
			Float absorption_g = (-log(color_g) - absorptionFactor * medulla_occlusion * absorption_inner) / (absorptionFactor - absorptionFactor * medulla_occlusion);
			Float absorption_b = (-log(color_b) - absorptionFactor * medulla_occlusion * absorption_inner) / (absorptionFactor - absorptionFactor * medulla_occlusion);

			absorption_r = clamp(absorption_r, 0.01, MAX_ABSORPTION_OUTER);
			absorption_g = clamp(absorption_g, 0.01, MAX_ABSORPTION_OUTER);
			absorption_b = clamp(absorption_b, 0.01, MAX_ABSORPTION_OUTER);

			absorption = Vector3(absorption_r, absorption_g, absorption_b);
		}
		//convert the w to theta and phi
		Vector3 wi = bRec.its.toWorld(bRec.wi);
		Vector3 wo = bRec.its.toWorld(bRec.wo);
		const Vector3 &u = bRec.its.geoFrame.s;
		const Vector3 &v = bRec.its.geoFrame.t;
		const Vector3 &w = bRec.its.geoFrame.n;

		Float theta_i = M_PI / 2.0f - acosSafe(dot(u, wi));
		Float theta_r = M_PI / 2.0f - acosSafe(dot(u, wo));
		//rolling the direction on the sphere
		regularizetheta(theta_i, phi_i);
		regularizetheta(theta_r, phi_r);
		Float phi_i = acosSafe(dot(normalizeSafe(wi - dot(u, wi) * u), v)) * (dot(w, wi) > 0 ? 1 : -1);
		Float phi_r = acosSafe(dot(normalizeSafe(wo - dot(u, wo) * u), v)) * (dot(w, wo) > 0 ? 1 : -1);
		Float phi = phi_r - phi_i;
		regularizePhi(phi);
		Float rho = bRec.its.storage[7];//geometery parameter of the aggregated fiber
		int numoffibers = bRec.its.storage[8];//geometery parameter of the aggregated fiber
		AggregatedSolver AggregatedSolver(eta, absorption, alpha, beta_m, beta_n,
			medulla_ratio, absorption_inner, scattering_inner, g, layers, rho, numoffibers, theta_i, bRec.sampler);
		Float h;		
		h = (AggregatedSolver.randomnum() * 2 - 1) * (1.0 - EPS);//h for the inhrited part
		return std::max(Float(0.0f), AggregatedSolver.pdfMN(h, theta_i, theta_r, phi));
	}

	// Choose a lobe and sample it.
	Spectrum sample(BSDFSamplingRecord &bRec, Float &pdf, const Point2 &sample) const {
		Vector3 absorption = absorption_outer;
		Float umin = bRec.its.storage[3];//the color area covered by an aggregated fiber
		Float umax = bRec.its.storage[4];//the color area covered by an aggregated fiber
		Float vmin = bRec.its.storage[5];//the color area covered by an aggregated fiber
		Float vmax = bRec.its.storage[6];//the color area covered by an aggregated fiber
		if (useColorTexture && medulla_ratio < 1) {
			Spectrum color = averagecolor(colorTexture, umin, umax, vmin, vmax);//average the absorption color
			Float color_r, color_g, color_b;
			color.toLinearRGB(color_r, color_g, color_b);

			color_r = clamp(std::pow(color_r, Float(1.0 / colorGamma)) * colorScale, Float(1.0 / 256.0), Float(0.99));
			color_g = clamp(std::pow(color_g, Float(1.0 / colorGamma)) * colorScale, Float(1.0 / 256.0), Float(0.99));
			color_b = clamp(std::pow(color_b, Float(1.0 / colorGamma)) * colorScale, Float(1.0 / 256.0), Float(0.99));

			Float medulla_occlusion = std::pow(medulla_ratio, 1.5);
			Float absorption_r = (-log(color_r) - absorptionFactor * medulla_occlusion * absorption_inner) / (absorptionFactor - absorptionFactor * medulla_occlusion);
			Float absorption_g = (-log(color_g) - absorptionFactor * medulla_occlusion * absorption_inner) / (absorptionFactor - absorptionFactor * medulla_occlusion);
			Float absorption_b = (-log(color_b) - absorptionFactor * medulla_occlusion * absorption_inner) / (absorptionFactor - absorptionFactor * medulla_occlusion);
			absorption_r = clamp(absorption_r, 0.01, MAX_ABSORPTION_OUTER);
			absorption_g = clamp(absorption_g, 0.01, MAX_ABSORPTION_OUTER);
			absorption_b = clamp(absorption_b, 0.01, MAX_ABSORPTION_OUTER);
			absorption = Vector3(absorption_r, absorption_g, absorption_b);
		}
		//convert the w to theta and phi
		Vector3 wi = bRec.its.toWorld(bRec.wi);		
		const Vector3 &u = bRec.its.geoFrame.s;
		const Vector3 &v = bRec.its.geoFrame.t;
		const Vector3 &w = bRec.its.geoFrame.n;
		//rolling the incident direction on the sphere~~
		Float theta_i = M_PI / 2.0f - acosSafe(dot(u, wi));
		regularizetheta(theta_i, phi_i);
		Float phi_i = acosSafe(dot(normalizeSafe(wi - dot(u, wi) * u), v)) * (dot(w, wi) > 0 ? 1 : -1);
		Float rho = bRec.its.storage[7];//geometery parameter of the aggregated fiber
		int numoffibers = bRec.its.storage[8];//geometery parameter of the aggregated fiber
		AggregatedSolver AggregatedSolver(eta, absorption, alpha, beta_m, beta_n,
			medulla_ratio, absorption_inner, scattering_inner, g, layers, rho, numoffibers, theta_i, bRec.sampler);

		// Choose a lobe to sample.
		Float h;
		h = (AggregatedSolver.randomnum() * 2 - 1) * (1.0 - EPS);//h for the inherited part
		int lobe;
		Float pdfLobe;
		AggregatedSolver.chooseLobe(theta_i, h, lobe, pdfLobe);
		bRec.sampledComponent = lobe < 3 ? 0 : 1;
		if (lobe == -1) {
			pdf = 0.0f;
			return Spectrum(0.0f);
		}
		if (((furMaskCode >> lobe) & 1) == 0) {
			pdf = 0.0f;
			return Spectrum(0.0f);
		}
		// Sample the chosen lobe.
		Float theta_o, phi_o;
		Float pdfM, pdfN;
		Float weightM;
		Vector3 weightN;
		AggregatedSolver.sampleLobe(lobe, h, theta_i, phi_i, theta_o, pdfM, weightM, phi_o, pdfN, weightN);
		//rolling on the sphere
		bool turnto = turntheta(theta_o);
		if (turnto)
		{
			phi_o += M_PI;
			if (theta_o > M_PI / 2.0f)
			{
				theta_o = M_PI - theta_o;
			}
			if (theta_o <-M_PI / 2.0f)
			{
				theta_o = -M_PI - theta_o;
			}
		}
		regularizePhi(phi_o);
		regularizeTheta(theta_o);
		theta_o = clamp(theta_o, degToRad(-89.5), degToRad(89.5));
		// Convert angles to outgoing direction.
		Float ul = sin(theta_o);
		Float vl = cos(theta_o) * cos(phi_o);
		Float wl = cos(theta_o) * sin(phi_o);
		bRec.wo = bRec.its.toLocal(u * ul + v * vl + w * wl);
		// Calculate the final sampling weight and pdf.
		Vector3 weight = weightM * weightN / pdfLobe;
		if (std::isnan(weight[0]) || std::isnan(weight[1]) || std::isnan(weight[2]))
			return Spectrum(0.0f);
		if (std::isinf(weight[0]) || std::isinf(weight[1]) || std::isinf(weight[2]))
			return Spectrum(0.0f);
		Float phi = phi_o - phi_i;
		pdf = pdfM * pdfN * pdfLobe / cos(theta_o);
		Spectrum retVal(0.f);
		retVal.fromLinearRGB(weight[0], weight[1], weight[2]);
		return retVal;
	}

	Spectrum sample(BSDFSamplingRecord &bRec, const Point2 &sample) const {
		Float pdf;
		return aggregated::sample(bRec, pdf, sample);
	}

	void addChild(const std::string &name, ConfigurableObject *child) {
		if (child->getClass()->derivesFrom(MTS_CLASS(Texture))
			&& name == "color") {
			colorTexture = static_cast<Texture *>(child);
		}
		else if (child->getClass()->derivesFrom(MTS_CLASS(Sensor))) {
			perspectiveCamera = static_cast<PerspectiveCamera *>(child);
		}
		else {
			BSDF::addChild(name, child);
		}
	}

	void serialize(Stream *stream, InstanceManager *manager) const {
		BSDF::serialize(stream, manager);
		manager->serialize(stream, colorTexture.get());
		//        perspectiveCamera->serialize(stream, manager);
		stream->writeBool(useColorTexture);
		stream->writeFloat(medulla_ratio);
		stream->writeFloat(eta);
		stream->writeFloat(beta_m);
		stream->writeFloat(beta_n);
		stream->writeFloat(alpha);
		stream->writeFloat(g);
		stream->writeFloat(layers);
		stream->writeFloat(absorptionFactor);
		stream->writeFloat(absorption_outer.x);
		stream->writeFloat(absorption_outer.y);
		stream->writeFloat(absorption_outer.z);
		stream->writeFloat(absorption_inner);
		stream->writeFloat(scattering_inner);
		stream->writeString(renderMode);
		stream->writeFloat(colorScale);
		stream->writeFloat(colorGamma);
	}

	Float getRoughness(const Intersection &its, int component) const {
		return std::numeric_limits<Float>::infinity();
	}

	std::string toString() const {
		std::ostringstream oss;
		oss << "SmoothDiffuse[" << endl
			<< "  id = \"" << getID() << "\"," << endl
			<< "  reflectance = " << indent(colorTexture->toString()) << endl
			<< "]";
		return oss.str();
	}

	Shader *createShader(Renderer *renderer) const;

	MTS_DECLARE_CLASS()
private:
	std::string renderMode;
	Float colorScale, colorGamma;
	ref<Texture> colorTexture;
	bool useColorTexture;
	Float medulla_ratio;
	Float eta;
	Vector3 absorption_outer;
	Float beta_m, beta_n;
	Float scattering_inner, absorption_inner;
	Float alpha;
	Float g;
	Float layers;
	Float absorptionFactor;
	Float directtrans;
	Float hairnum;
	Float radius;
	int furMaskCode;
	PerspectiveCamera *perspectiveCamera;
};

// ================ Hardware shader implementation ================

class aggregatedShader : public Shader {
public:
	aggregatedShader(Renderer *renderer, const Texture *reflectance)
		: Shader(renderer, EBSDFShader), colorTexture(reflectance) {
		colorTextureShader = renderer->registerShaderForResource(colorTexture.get());
	}

	bool isComplete() const {
		return colorTextureShader.get() != NULL;
	}

	void cleanup(Renderer *renderer) {
		renderer->unregisterShaderForResource(colorTexture.get());
	}

	void putDependencies(std::vector<Shader *> &deps) {
		deps.push_back(colorTextureShader.get());
	}

	void generateCode(std::ostringstream &oss,
		const std::string &evalName,
		const std::vector<std::string> &depNames) const {
		oss << "vec3 " << evalName << "(vec2 uv, vec3 wi, vec3 wo) {" << endl
			<< "    if (cosTheta(wi) < 0.0 || cosTheta(wo) < 0.0)" << endl
			<< "        return vec3(0.0);" << endl
			<< "    return " << depNames[0] << "(uv) * inv_pi * cosTheta(wo);" << endl
			<< "}" << endl
			<< endl
			<< "vec3 " << evalName << "_diffuse(vec2 uv, vec3 wi, vec3 wo) {" << endl
			<< "    return " << evalName << "(uv, wi, wo);" << endl
			<< "}" << endl;
	}

	MTS_DECLARE_CLASS()
private:
	ref<const Texture> colorTexture;
	ref<Shader> colorTextureShader;
};

Shader *aggregated::createShader(Renderer *renderer) const {
	return new aggregatedShader(renderer, colorTexture.get());
}

MTS_IMPLEMENT_CLASS(aggregatedShader, false, Shader)
MTS_IMPLEMENT_CLASS_S(aggregated, false, BSDF)
MTS_EXPORT_PLUGIN(aggregated, "Our aggregated BSDF")
MTS_NAMESPACE_END
